import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import com.mindfusion.common.DateTime;
import com.mindfusion.drawing.Colors;
import com.mindfusion.spreadsheet.Workbook;
import com.mindfusion.spreadsheet.Worksheet;
import com.mindfusion.spreadsheet.WorksheetView;
import com.mindfusion.spreadsheet.expressions.ErrorType;
import com.mindfusion.spreadsheet.expressions.IExpressionEvaluator;
import com.mindfusion.spreadsheet.expressions.IExpressionEvaluatorContext;
import com.mindfusion.spreadsheet.expressions.IInformationProvider;
import com.mindfusion.spreadsheet.expressions.XmlInformationProvider;


public class CustomFunctions
{
	public static void main(String[] args)
	{			
		SwingUtilities.invokeLater(new Runnable()
		{
			public void run()
			{
				MainWindow window = new MainWindow();
				window.setTitle("JSpreadsheet Sample: Custom Funcions");
				window.setVisible(true);
			}
		});
	}
	
	public static class MainWindow extends JFrame
	{
		public MainWindow()
		{
			setDefaultCloseOperation(EXIT_ON_CLOSE);
			setSize(1280, 1024);

			Workbook workbook = new Workbook();
			Worksheet activeSheet = workbook.getWorksheets().add();

			WorksheetView view = new WorksheetView();
			view.setWorksheet(activeSheet);
			add(view);

			workbook.setExtendedEvaluator(new CustomFunctionsEvaluator());

			IInformationProvider infoProvider = null;
			InputStream stream = CustomFunctions.class.getResourceAsStream("CustomFunctionInformation.xml");
			if (stream == null)
				return;

			try
			{
				DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			    Document document = builder.parse(stream);
			    infoProvider = XmlInformationProvider.create(document);
				stream.close();
			}
			catch (ParserConfigurationException | SAXException | IOException e)
			{
				System.out.println(e.getMessage());
			}
			workbook.setExtendedInformationProvider(infoProvider);

			activeSheet.beginInit();
			activeSheet.getCells().get(1, 0).setData("Result");
			activeSheet.getCells().get(2, 0).setData("Formula");
			activeSheet.getCells().get(3, 0).setData("Description");
			activeSheet.getCells().get(1, 1).setData("=EASTER(YEAR(TODAY()))");
			activeSheet.getCells().get(2, 1).setData("'=EASTER(YEAR(TODAY()))");
			activeSheet.getCells().get(3, 1).setData(infoProvider.getFunctionSummary("EASTER"));
			activeSheet.getCells().get(1, 2).setData("=CLENGTH(10)");
			activeSheet.getCells().get(2, 2).setData("'=CLENGTH(10)");
			activeSheet.getCells().get(3, 2).setData(infoProvider.getFunctionSummary("CLENGTH"));
			activeSheet.getCells().get(1, 3).setData("=CAREA(10)");
			activeSheet.getCells().get(2, 3).setData("'=CAREA(10)");
			activeSheet.getCells().get(3, 3).setData(infoProvider.getFunctionSummary("CAREA"));
			activeSheet.getCellRanges().get(1, 0, 3, 0).getStyle().setTextColor(Colors.Gray);
			activeSheet.endInit();
			view.resizeColumnsToFit(1, 3);
		}

		public static class CustomFunctionsEvaluator implements IExpressionEvaluator
		{
			public Object evaluateIdentifier(IExpressionEvaluatorContext context, String identifier)
			{
				return null;
			}

			public Object invokeFunction(IExpressionEvaluatorContext context, String name, List<?> parameters)
			{
				switch (name)
				{

					case "EASTER":
					{
						if (parameters.size() != 1)
							return context.getError(ErrorType.Value, "Wrong number of arguments.");
	
						Double year = context.toDouble(parameters.get(0));
						if (year == null)
							return context.getError(ErrorType.Value, "Wrong argument type.");
	
						return easterSundayOf(year.intValue());
					}

					case "CLENGTH":
					{
						if (parameters.size() != 1)
							return context.getError(ErrorType.Value, "Wrong number of arguments.");
	
						Double radius = context.toDouble(parameters.get(0));
						if (radius == null)
							return context.getError(ErrorType.Value, "Wrong argument type.");
	
						return 2 * Math.PI * radius.doubleValue();
					}

					case "CAREA":
					{
						if (parameters.size() != 1)
							return context.getError(ErrorType.Value, "Wrong number of arguments.");
	
						Double radius = context.toDouble(parameters.get(0));
						if (radius == null)
							return context.getError(ErrorType.Value, "Wrong argument type.");
	
						// Note: obtain PI by calling the respective function of the core evaluator
						Double pi = context.toDouble(context.getCoreEvaluator().invokeFunction(context, "PI", Arrays.asList()));
						if (pi == null)
							return context.getError(ErrorType.Value, "Cannot evaluate PI.");

						return pi.doubleValue() * Math.pow(radius.doubleValue(), 2);
					}

					default:
						break;
				}
				

				return null;
			}

			private static DateTime easterSundayOf(int year)
			{
				int Y = year;
				int a = Y % 19;
				int b = Y / 100;
				int c = Y % 100;
				int d = b / 4;
				int e = b % 4;
				int f = (b + 8) / 25;
				int g = (b - f + 1) / 3;
				int h = (19 * a + b - d - g + 15) % 30;
				int i = c / 4;
				int k = c % 4;
				int L = (32 + 2 * e + 2 * i - h - k) % 7;
				int m = (a + 11 * h + 22 * L) / 451;
				int Month = (h + L - 7 * m + 114) / 31;
				int Day = ((h + L - 7 * m + 114) % 31) + 1;
				return new DateTime(year, Month, Day);
			}
		}


		private static final long serialVersionUID = 1L;
	}
}